"""
This type stub file was generated by pyright.
"""

"""Benchmarks for utilities working with arbitrarily nested structures."""
TIME_UNITS = ...
def format_time(time): # -> str | None:
  ...

def run_benchmark(benchmark_fn, num_iters):
  ...

def map_to_list(func, *args): # -> list[Any]:
  ...

def benchmark_map(map_fn, structure): # -> Callable[[], Any]:
  ...

BENCHMARKS = ...
def main(): # -> None:
  ...

if __name__ == "__main__":
  ...
