@import url('https://fonts.googleapis.com/css2?family=Public+Sans:wght@400;700&display=swap');
@import url('https://fonts.googleapis.com/css2?family=Quicksand:wght@300..700&display=swap');

body {
    font-family: "Quicksand", sans-serif !important; 
}

h1 {
    font-family: 'Public Sans', sans-serif !important;
    font-weight: 700 !important;
}

h2 {
    font-family: 'Public Sans', sans-serif !important;
    font-weight: 600 !important;
}

h3 {
    font-family: 'Public Sans', sans-serif !important;
    font-weight: 500 !important;
}

h4 {
    font-family: 'Public Sans', sans-serif !important;
    font-weight: 400 !important;
}